// generate text from a file

#include "globals.hh"
#include <iostream>
#include <stdlib.h>
#include <time.h>

hash_map<deque<SymbolID>, vector<pair<SymbolID, float> >, deque_hash> probabilities;

// random number generator state vector
char randomstate[256];

// random number generator (uniformly distributed)
void initrandom()
{
  unsigned int seed = time(NULL);
 
  cout << "using random number generator seed: " << seed << endl;
 
  // use a good random number generator
  initstate(seed, randomstate, 256);
  setstate(randomstate);
}

// generates a symbol, using the probabilities map
Symbol generateSymbol(deque<SymbolID> const &memory)
{
  // get a random number between 0 and 1
  float r = random()/(float(RAND_MAX)+1);

  vector<pair<SymbolID, float> > const &prob = probabilities[memory];
  
  // convert random number to Symbol ID according to distribution
  for (vector<pair<SymbolID, float> >::const_iterator i = prob.begin(); 
       i != prob.end(); i++)
    {
      r -= i->second;

      // if r was between [sum up to i-1] and [sum up to i], this is 
      // what we want.
      if (r < 0)
	return id2symbol[i->first];
    }

  // nothing here, vector empty (maybe eof while reading in): we
  // should return a symbol according to the absolute probabilities
  // (which we don't have).
  // Just return one which is uniformly distributed.
  // We could also stop here.
  return id2symbol[int(id2symbol.size()*r)];
}

void readProbabilitiesChar(unsigned int order, istream &is)
{
  cout << "reading probabilities (char mode)..." << endl;

  // ignore rest of first line
  is.ignore(0x7fffffff, '\n');

  do
    {
      // ignore until end of line
      is.ignore(0x7fffffff, '\n');
      
      // if this was the last line, break (why is this necessary?)
      if (is.eof())
	break;

      // read the combination (build memory)
      deque<SymbolID> memory;
      for (unsigned int i = 0; i < order; i++)
	{
	  Symbol s = char2symbol(is.get());
	  SymbolID sid = getSymbolID(s);
	  
	  memory.push_back(sid);
	}

      // get actual data
      Symbol s = char2symbol(is.get());
      SymbolID sid = getSymbolID(s);
      float prob;
      is >> prob;
      probabilities[memory].push_back(make_pair(sid, prob));
    }
  while(true);

  cout << "done, " << id2symbol.size() << " symbols, " 
       << probabilities.size() << " transitions." << endl;
}

void readProbabilitiesDynamic(unsigned int order, istream &is)
{
  cout << "reading probabilities (dynamic mode)..." << endl;

  // read number of ids
  unsigned int numids;
  is >> numids;
  id2symbol.resize(numids);

  // read id to symbol table
  for (unsigned int i = 0; i < numids; i++)
    {
      SymbolID sid;
      Symbol s;
      
      is >> sid;
      is >> skipws >> s;

      id2symbol[sid] = s;
      symbolid[s] = sid;
    }

  cout << "symbols done, " << id2symbol.size() << " symbols." << endl;
  
  // read actual data
  do
    {
      // ignore until end of line
      is.ignore(0x7fffffff, '\n');

      if (is.eof())
	break;

      SymbolID sid;
      float prob;

      deque<SymbolID> memory;

      // fill memory
      for (unsigned int i = 0; i < order; i++)
	{
	  is >> sid;
	  memory.push_back(sid);
	}
      
      // get data
      is >> sid;
      is >> prob;

      probabilities[memory].push_back(make_pair(sid, prob));      
    }
  while (true);

  cout << "done, " << probabilities.size() << " transitions." << endl;
}

int generate(unsigned int n, istream &is, ostream &os)
{
  cout << "Generating " << n << " symbols." << endl;

  // read file type info from file
  string type;
  unsigned int order;

  is >> type >> order;

  cout << "input information type: " << type 
       << ", order " << order << "." << endl;
  
  initrandom();

  // deprecated
  if (type == "char")
    readProbabilitiesChar(order, is);
  // ahhh.
  else if (type == "dynamic")
    readProbabilitiesDynamic(order, is);

  deque<SymbolID> memory;

  // seed memory with a (more or less) random existing sequence
  float r = random()/(float(RAND_MAX)+1);
  unsigned int pos = int(probabilities.size()*r);
      
  // find the pos'th element of the probability table
  hash_map<deque<SymbolID>, vector<pair<SymbolID, float> >, deque_hash>::const_iterator iter = probabilities.begin();
  for (unsigned int i = 0; i < pos; i++)
    iter++;
  memory = iter->first;

  // push garbage in front (will be removed first thing in the loop)
  memory.push_front(0);

  // generate the words
  for (unsigned int i = 0; i < n; i++)
    {
      // remove old symbol
      memory.pop_front();

      // generate new symbol
      Symbol s = generateSymbol(memory);
      SymbolID sid = getSymbolID(s);

      // store new symbol
      memory.push_back(sid);

      // output new symbol
      cout << s << ' ';
      os << s << ' ';
    }

  return 0;
}
